/* Copyright (c) 2012 Robert Nordier.  All rights reserved. */

#include "oc+.h"

/*
 * Byte 0: Unused here
 * Byte 1: Pops from stack (0, 1, 2)
 * Byte 2: Pushes onto stack (0, 1, 2 = set to n, 3 = set to n + 1)
 * Byte 3: 0-7: force; 1-7: loadreg; 8: load op; 9: data op
 */

const char optab[][OPATTR] = {
    {0, 0, 0, 0},       /* ... */
    {0, 0, 0, 0},       /* ... */
    {0, 0, 0, 0},       /* ... */
    {0, 0, 0, 0},       /* ... */
    {1, 0, 1, 8},   /* S_TRUE 4 */
    {1, 0, 1, 8},   /* S_FALSE 5 */
    {0, 0, 0, 0},       /* ... */
    {0, 0, 0, 0},       /* ... */
    {1, 1, 1, 3},   /* S_RV 8 */
    {0, 0, 0, 0},       /* ... */
    {2, 1, 3, 1},   /* S_FNAP 10 */
    {1, 0, 1, 8},   /* S_QUERY 11 */
    {1, 1, 1, 3},   /* S_NEG 12 */
    {1, 1, 1, 3},   /* S_ABS 13 */
    {1, 2, 1, 5},   /* S_MULT 14 */
    {1, 2, 1, 5},   /* S_DIV 15 */
    {1, 2, 1, 5},   /* S_REM 16 */
    {1, 2, 1, 4},   /* S_PLUS 17 */
    {1, 2, 1, 4},   /* S_MINUS 18*/
    {1, 2, 1, 4},   /* S_EQ 19 */
    {1, 2, 1, 4},   /* S_NE 20 */
    {1, 2, 1, 4},   /* S_LS 21 */
    {1, 2, 1, 4},   /* S_GR 22 */
    {1, 2, 1, 4},   /* S_LE 23 */
    {1, 2, 1, 4},   /* S_GE 24 */
    {0, 0, 0, 0},       /* (S_SLCT) */
    {0, 0, 0, 0},       /* (S_OF) */
    {0, 0, 0, 0},       /* (S_BYTEAP) */
    {0, 0, 0, 0},       /* (S_MTHAP) */
    {1, 1, 1, 3},   /* S_NOT 29 */
    {1, 2, 1, 6},   /* S_LSHIFT 30 */
    {1, 2, 1, 6},   /* S_RSHIFT 31 */
    {1, 2, 1, 4},   /* S_LOGAND 32 */
    {1, 2, 1, 4},   /* S_LOGOR 33 */
    {1, 2, 1, 4},   /* S_EQV 34 */
    {1, 2, 1, 4},   /* S_NEQV 35 */
    {0, 0, 0, 0},       /* (S_COND) */
    {0, 0, 0, 0},       /* (S_COMMA) */
    {0, 0, 0, 0},       /* (S_TABLE) */
    {4, 0, 0, 0},       /* S_NEEDS 39 */
    {4, 0, 0, 0},   /* S_SECTION 40 */
    {0, 0, 0, 0},       /* (S_ASS) */
    {2, 1, 2, 1},   /* S_RTAP 42 */
    {1, 1, 0, 1},   /* S_GOTO 43 */
    {0, 0, 0, 0},       /* (S_RESULTIS) */
    {0, 0, 0, 0},       /* (S_COLON) */
    {0, 0, 0, 0},       /* (S_TEST) */
    {0, 0, 0, 0},       /* (S_FOR) */
    {0, 0, 0, 0},       /* (S_IF) */
    {0, 0, 0, 0},       /* (S_UNLESS) */
    {0, 0, 0, 0},       /* (S_WHILE) */
    {0, 0, 0, 0},       /* (S_UNTIL) */
    {0, 0, 0, 0},       /* (S_REPEAT) */
    {0, 0, 0, 0},       /* (S_REPEATWHILE) */
    {0, 0, 0, 0},       /* (S_REPEATUNTIL) */
    {0, 0, 0, 0},       /* (S_SKIP) */
    {0, 0, 0, 0},       /* (S_LOOP) */
    {0, 0, 0, 0},       /* (S_BREAK) */
    {1, 0, 0, 0},   /* S_RETURN 58 */
    {1, 0, 0, 0},   /* S_FINISH 59 */
    {0, 0, 0, 0},       /* (S_ENDCASE) */
    {7, 1, 0, 3},   /* S_SWITCHON 61 */
    {0, 0, 0, 0},       /* (S_CASE) */
    {0, 0, 0, 0},       /* (S_DEFAULT) */
    {0, 0, 0, 0},       /* (S_SEQ) */
    {0, 0, 0, 0},       /* (S_LET) */
    {0, 0, 0, 0},       /* (S_AND) */
    {0, 0, 0, 0},       /* (S_MANIFEST) */
    {6, 0, 0, 0},   /* S_GLOBAL 68 */
    {0, 0, 0, 0},       /* (S_STATIC) */
    {0, 0, 0, 0},       /* (S_VALDEF) */
    {0, 0, 0, 0},       /* (S_VECDEF) */
    {0, 0, 0, 0},       /* (S_CONSTDEF) */
    {0, 0, 0, 0},       /* (S_CONST) */
    {0, 0, 0, 0},       /* (S_FNDEF) */
    {0, 0, 0, 0},       /* (S_RTDEF) */
    {0, 0, 0, 0},       /* (S_LOCAL) */
    {0, 0, 0, 0},       /* (S_LABEL) */
    {0, 0, 0, 0},       /* (S_BE) */
    {0, 0, 0, 0},       /* (S_LSECT) */
    {0, 0, 0, 0},       /* (S_RSECT) */
    {0, 0, 0, 0},       /* (S_GET) */
    {0, 0, 0, 0},       /* (S_SEMICOLON) */
    {0, 0, 0, 0},       /* (S_INTO) */
    {0, 0, 0, 0},       /* (S_TO) */
    {0, 0, 0, 0},       /* (S_By) */
    {0, 0, 0, 0},       /* (S_DO) */
    {0, 0, 0, 0},       /* (S_ELSE) */
    {0, 0, 0, 0},       /* (S_VEC) */
    {0, 0, 0, 0},       /* (S_LPAREN) */
    {0, 0, 0, 0},       /* (S_RPAREN) */
    {0, 0, 0, 0},       /* (S_SBRA) */
    {0, 0, 0, 0},       /* (S_SKET) */
    {0, 0, 0, 0},       /* (S_DOT) */
    {0, 0, 0, 0},       /* (S_EOF) */
    {0, 0, 0, 0},       /* (S_NONE) */
    {3, 0, 1, 8},   /* S_LF 96 */
    {2, 0, 1, 8},   /* S_LP 97 */
    {2, 0, 1, 8},   /* S_LG 98*/
    {2, 0, 1, 8},   /* S_LN 99 */
    {4, 0, 1, 8},   /* S_LSTR 100 */
    {3, 0, 1, 8},   /* S_LL 101*/
    {2, 0, 1, 8},   /* S_LLP 102*/
    {2, 0, 1, 8},   /* S_LLG 103*/
    {3, 0, 1, 8},   /* S_LLL 104*/
    {2, 1, 0, 2},   /* S_SP 105 */
    {2, 1, 0, 2},   /* S_SG 106 */
    {3, 1, 0, 2},   /* S_SL 107 */
    {1, 2, 0, 6},   /* S_STIND 108 */
    {3, 0, 0, 0},   /* S_JUMP 109 */
    {3, 1, 0, 3},   /* S_JT 110 */
    {3, 1, 0, 3},   /* S_JF 111 */
    {3, 2, 0, 4},   /* S_ENDFOR 112 */
    {3, 0, 0, 0},   /* S_LAB 113 */
    {2, 0, 2, 0},   /* S_STACK 114 */
    {1, 0, 0, 0},   /* S_STORE 115 */
    {2, 0, 3, 0},   /* S_RSTACK 116 */
    {5, 0, 0, 0},   /* S_ENTRY 117 */
    {2, 0, 2, 0},   /* S_SAVE 118 */
    {1, 1, 0, 3},   /* S_FNRN 119 */
    {1, 0, 0, 0},   /* S_RTRN 120 */
    {3, 1, 0, 3},   /* S_RES 121 */
    {3, 0, 0, 9},   /* S_DATALAB 122 */
    {2, 0, 0, 9},   /* S_ITEMN 123 */
    {2, 0, 0, 0},   /* S_ENDPROC 124 */
    {1, 2, 1, 4},   /* S_GETBYTE 125 */
    {1, 2, 0, 4},   /* S_PUTBYTE 126 (actually 3 args) */
    {1, 0, 0, 0},   /* S_END 127 */
    {0, 0, 0, 0},       /* (S_DEBUG) */
    {3, 0, 0, 9},   /* S_ITEML 129 */
    {0, 0, 0, 0}        /* (S_CHAR) */
};
