// This is an OCODE to Intcode codegenerator dated 1982 approx

//    CGHDR


// GET "LIBHDR"

// OCODE OPERATORS, 1982 compiler style:
/*
MANIFEST
$(    s.true          =    4
      s.false         =    5
      s.rv            =    8
      s.fnap          =   10
      s.mult          =   11
      s.div           =   12
      s.rem           =   13
      s.plus          =   14
      s.minus         =   15
      s.query         =   16
      s.neg           =   17
      s.abs           =   19
      s.eq            =   20
      s.ne            =   21
      s.ls            =   22
      s.gr            =   23
      s.le            =   24
      s.ge            =   25
      s.not           =   30
      s.lshift        =   31
      s.rshift        =   32
      s.logand        =   33
      s.logor         =   34
      s.eqv           =   35
      s.neqv          =   36
      s.cond          =   37
      s.lp            =   40
      s.lg            =   41
      s.ln            =   42
      s.lstr          =   43
      s.ll            =   44
      s.llp           =   45
      s.llg           =   46
      s.lll           =   47
      s.needs         =   48
      s.section       =   49
      s.rtap          =   51
      s.goto          =   52
      s.for           =   56
      s.return        =   67
      s.finish        =   68
      s.switchon      =   70
      s.global        =   76
      s.sp            =   80
      s.sg            =   81
      s.sl            =   82
      s.stind         =   83
      s.jump          =   85
      s.jt            =   86
      s.jf            =   87
      s.endfor        =   88
      s.lab           =   90
      s.stack         =   91
      s.store         =   92
      s.rstack        =   93
      s.entry         =   94
      s.save          =   95
      s.fnrn          =   96
      s.rtrn          =   97
      s.res           =   98
      s.reslab        =   99
      s.datalab       =  100
      s.iteml         =  101
      s.itemn         =  102
      s.endproc       =  103
      s.end           =  104
      s.char          =  105
      error           =  108
      s.debug         =  109
      s.startblock    =  110
      s.none          =  111
      s.mod           =  116
      s.getbyte       =  120
      s.putbyte       =  121
$)
*/

// BCPL lexical tokens, tree and ocode operators, 2010 compiler style:

MANIFEST $(
s.number=1
s.name; s.string; s.true; s.false
s.valof; s.lv; s.rv; s.vecap; s.fnap						// 10
s.query
s.neg; s.abs // Integer operators
s.mult; s.div; s.rem; s.plus; s.minus
s.eq; s.ne; s.ls; s.gr; s.le; s.ge							// 24
s.slct; s.of                   // Inserted 11/7/01
s.byteap; s.mthap
s.not; s.lshift; s.rshift; s.logand; s.logor
s.eqv; s.neqv; s.cond; s.comma; s.table						// 38
s.needs; s.section
s.ass
s.rtap; s.goto; s.resultis; s.colon
s.test; s.for; s.if; s.unless								// 49
s.while; s.until; s.repeat; s.repeatwhile; s.repeatuntil
s.skip // Added 22/6/05
s.loop; s.break; s.return; s.finish
s.endcase; s.switchon; s.case; s.default					// 63
s.seq; s.let; s.and; s.manifest; s.global; s.static
s.valdef; s.vecdef; s.constdef; s.const
s.fndef; s.rtdef; s.local; s.label							// 77

// Othe lexical token
s.be; s.lsect; s.rsect; s.get
s.semicolon; s.into; s.to; s.by; s.do; s.else
s.vec; s.lparen; s.rparen; s.sbra; s.sket; s.dot; s.eof		// 94
s.bitsperbcplword

// Used in the code generators
s.none														// 96

// Ocode operators
s.lf; s.lp; s.lg; s.ln; s.lstr; s.ll; s.llp; s.llg; s.lll 
s.sp; s.sg; s.sl; s.stind; s.jump; s.jt; s.jf; s.endfor		// 113
s.lab; s.stack; s.store; s.rstack; s.entry
s.save; s.fnrn; s.rtrn; s.res; s.datalab; s.itemn
s.endproc; s.getbyte; s.putbyte
s.end; s.debug; s.iteml; s.char								// 131

// The following have been added for the extended compiler xbcpl

// Floating point operators and assignment operators, added 15/07/10

s.fnum // Floating point constants							// 132
s.float; s.fix; s.fabs
s.fmul; s.fdiv; s.fadd; s.fsub;  s.fpos; s.fneg
s.feq; s.fne; s.fls; s.fgr; s.fle; s.fge					// 147

// Assign operators -- added 15/07/10
s.assvecap
s.assmul; s.assdiv; s.assrem; s.assadd; s.asssub			// 153
s.assfmul; s.assfdiv; s.assfadd; s.assfsub
s.asslshift; s.assrshift
s.asslogand; s.asslogor; s.asseqv; s.assneqv				// 163


s.selld; s.selst // Added 19/07/10

s.fltop  // FLTOP is followed by one of the fl. codes		// 166
         // eg FLTOP FADD to do a := b #+ a
         // or FLTOP FLOAT to do a := FLOAT a

sf.none=0     // Assignment operators						// 167
sf.vecap
sf.fmul
sf.fdiv
sf.fadd
sf.fsub														// 172
sf.mul
sf.div
sf.rem
sf.add
sf.sub
sf.lshift													// 178
sf.rshift
sf.logand
sf.logor
sf.eqv
sf.neqv														// 183
$)




// Code Generator Globals:

GLOBAL
$(  rc:                  150
    fromfile:            151
    verstream:           152
    ocodestream:         153
    codestream:          154
    datvec:              155
    err.p:               156
    err.l:               157
    ocodefile:           158
    keepocode:           159

    // CG options

    cgworksize:          160
    cglisting:           161
    naming:              162
    callcounting:        163
    profcounting:        164
    stkchking:           165
    restricted:          166
    altobj:              167
    cg.a:                168
    cg.b:                169
    cg.y:                175
    cg.z:                176
$)



MANIFEST
$(  free.global = 177
$)




GLOBAL
$(  ch:         free.global+0
    wordv:      free.global+1
    sysprint:   free.global+3
    sysin:      free.global+4
    ssp:        free.global+5
    state:      free.global+6
    ad.a:       free.global+7
    ad.k:       free.global+8
    option:     free.global+9
    datav:      free.global+10
    datap:      free.global+11
    datat:      free.global+12
    proglength: free.global+13
    linep:      free.global+14
    param:      free.global+15
    op:         free.global+16
    readop:     free.global+17
    rdn:        free.global+18
    rdl:        free.global+19
    gencode:    free.global+20
    force.nil:  free.global+21
    force.ad:   free.global+22
    force.ac:   free.global+23
    force.acad: free.global+24
    swap:       free.global+25
    load:       free.global+26
    storein:    free.global+27
    cgstring:   free.global+28
    data:       free.global+29
    nextparam:  free.global+30
    code:       free.global+31
    complab:    free.global+32
    opcode:     free.global+33
    wr:         free.global+34
    wrn:        free.global+35
    wrdata:     free.global+36
$)

MANIFEST $(
    m.n           = 0
    m.i           = 1
    m.p           = 2
    m.ip          = 3
    m.l           = 4
    m.il          = 5
    m.g           = 6
    m.ig          = 7
	m.f           = 8
	m.if          = 9
	
    f.l           = 'L'
    f.s           = 'S'
    f.a           = 'A'
    f.j           = 'J'
    f.t           = 'T'
    f.f           = 'F'
    f.k           = 'K'
    f.x           = 'X'
    f.d           = 'D'
    f.c           = 'C'
    nil           = 0
    ad            = 1
    ac            = 2
    acad          = 3
$)
