/*
MCPL to MIAL (MCPL Internal Assembly Language) compiler
for compilation into native code using, for example, mial386.

Copyright: Martin Richards 16 June 1997


It is based on the MCPL to MINTCODE and the BCPL to SIAL compilers.

It uses a reverse polish form of MCODE.

Development of this compiler started on 2 Nov 1992
This version is designed to run interpretively using MINTCODE,
and is similar in structure to the BCPL Cintcode system.
The GLOBAL declaration has been put back into the MCPL and
the syntax of EXTERNAL declarations has been retained but not
implemented.

9/7/97  Separate mcplsyn.b and mcpltrn.b from mcpl2mial.b
        so that they can be shared with the mcpl->mintcode compiler.
*/

GET "mcplsyn.b"
.


GET "mcpltrn_w.b"
.


GET "mcplcg.b"
.


SECTION "MCPL2MIAL"

GET "libhdr"

GLOBAL {
//General Interface
errcount:600; errmax:601; fin_p:602; fin_l:603
sourcestream:604; sysprint:605; mcodeout:606
treep:607; treevec:608; findfilename:609

//Syn interface
chbuf:610; ch:611; chcount:612; lineno:613
rch:614; findfileno:615; filelist:616; freefilelist:617
formtree:618; plist:619

// Trn interface
translate:620

// Cg interface
mcodein:630; tostream: 631
codegenerate: 632
wrarg: 633; codelayout:634
}

LET start() = VALOF
{  LET treesize = 200000
   AND argv = VEC 50
   AND argform =
   "FROM/A,TO/K,VER/K,TREE/S,D1/S,D2/S,OENDER/S,"
   LET debug, prtree, bigender = 0, FALSE, ?
   LET stdout = output()

   sysprint := stdout
   selectoutput(sysprint)

   writef("*nMCPL2MIAL 8 July 1997*n")

   IF rdargs(argform, argv, 50)=0 DO {  writes("Bad arguments*n")
                                        RESULTIS 20
                                     }

   fin_p, fin_l := level(), fin
   errmax       := 30
   errcount     := 0
   treevec      := 0
   sourcestream := 0
   mcodeout     := 0
   mcodein      := 0
   tostream     := 0
   filelist     := 0

   prtree := argv!3

   // Code generator options

   bigender := (!"AAA" & 255) = 'A' // =TRUE if running on a bigender
   IF argv!4 DO debug    := debug+1       // D1
   IF argv!5 DO debug    := debug+2       // D2
   IF argv!6 DO bigender := ~bigender     // OENDER

   sourcestream := findinput(argv!0)      // FROM

   IF sourcestream=0 DO {  writef("Trouble with FROM file %s*n", argv!0)
                           errcount := 1
                           GOTO fin
                        }

   selectinput(sourcestream)

   mcodeout := findoutput("MCODE")

   IF mcodeout=0 DO {  writes("Trouble with file MCODE*n")
                       errcount := 1
                       GOTO fin
                    }

   treevec := getvec(treesize)

   IF treevec=0 DO {  writes("Insufficient memory*n")
                      errcount := 1
                      GOTO fin
                   }

   UNLESS argv!2=0 DO       // VER
   {  sysprint := findoutput(argv!2)
      IF sysprint=0 DO
      {  sysprint := stdout
         writef("Trouble with VER file %s*n", argv!2)
         errcount := 1
         GOTO fin
      }
   }

   selectoutput(sysprint)

   {  LET b = VEC 64/bytesperword
      chbuf := b
      FOR i = 0 TO 63 DO chbuf%i := 0
      chcount, lineno := 0, findfileno(argv!0)<<24 | 1
      rch()

      UNTIL ch=endstreamch DO
      {  LET tree = ?
         treep := treevec + treesize

         tree := formtree()
         IF tree=0 BREAK

         writef("*nTree size %n*n", treesize+treevec-treep)

         IF prtree DO {  writes("Parse Tree*n")
                         plist(tree, 0, 20)
                         newline()
                      }

         UNLESS errcount=0 GOTO fin
         selectoutput(mcodeout)
         translate(tree)
         selectoutput(sysprint)
      }
   }

   selectinput(sourcestream);  endread();   sourcestream := 0
   selectoutput(mcodeout);     endwrite();  mcodeout     := 0

   selectoutput(sysprint)

   UNLESS errcount=0 GOTO fin

   mcodein := findinput("MCODE")
   IF mcodein=0 DO
   {  writef("Trouble reading file MCODE*n")
      errcount := 1
      GOTO fin
   }

   IF argv!1=0 DO argv!1 := "MIAL"

   tostream := findoutput(argv!1)
   IF tostream=0 DO
   {  writef("Trouble with code file %s*n", argv!1)
      errcount := 1
      GOTO fin
   }

   selectinput(mcodein)
   codegenerate(treevec, treesize, debug, bigender)

   selectoutput(sysprint)

fin:
   UNLESS treevec=0       DO freevec(treevec)
   UNLESS sourcestream=0  DO {  selectinput(sourcestream); endread()  }
   UNLESS mcodeout=0      DO {  selectoutput(mcodeout);    endwrite() }
   UNLESS mcodein=0       DO {  selectinput(mcodein);      endread()  }
   UNLESS tostream=0      DO {  selectoutput(tostream)
                                UNLESS tostream=stdout DO  endwrite() }
   UNLESS sysprint=stdout DO {  selectoutput(sysprint);    endwrite() }
   freefilelist(filelist)
   selectoutput(stdout)
   UNLESS errcount=0 RESULTIS 20
   RESULTIS 0
}




