@echo off

rem	Copyright (c) 2014,2015,2018 Dave Cannon.  All rights reserved.
rem	$Id: bcpl.cmd,v 1.03 7-Jul-2018 17:32:12 rn Exp $

setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

set vn=1
set rvn=03
set edt=7-Jul-2018

set version=%0 Vn %vn%.%rvn%	%edt%

rem	Vn 1.03	 7-Jul-2018 add cross-assemler option
rem	Vn 1.02	18-Oct-2015 minor bug fix
rem	Vn 1.01	20-Aug-2014 all environment variables now set externally by ..\BCPLenv.cmd
rem	Vn 1.00	 7-Aug-2014	Initial version for DOS under Windows (amended from the obcpl version)

if "%BCPLocd%"=="" (
  echo.
  echo 	'BCPLocd' - the port of Robert Nordier's OCODE-based assembler,
  echo 	BCPL-0.9.8, is required by this version of the BCPL compiler.
  echo 	Please download and install it.
  goto eof
)

if not '%1'=='' goto doit

:usage
echo	Usage:	%0 [-Options] [-o output] file
echo		where	file is a BCPL source program to be compiled
echo			[-o output] (optional) specifies the compiled program name
echo			[-x x-assembler] (optional) cross-assemble for named m/c
echo			[-Options] may be any of:
echo			[-O] (optional) suppresses the code optimiser
echo			[-c] (optional) preserve SIAL output as 'file.sial'
echo			[-l] (optional) preserve the 'file.o' object file for ld
echo			[-S] (optional) preserve the 'file.s' assembler file
echo.
echo		%version%
echo.
goto eof

:doit

set oflag=0
set args=

if not defined GnuWin (
  echo.
  echo 	The BCPL environment variables have not been established.
  echo 	You must execute ..\BCPLenv.cmd before running this script
  echo.
  goto eof
)

set LDFLAGS=-Bdynamic -T"%GnuWin%\mingw32\lib\ldscripts\i386pe_mgw.x" "%GnuWin%\lib\crt2.o" "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtbegin.o"
set LDLIBS=-lmingw32 -lmoldname -lmingwex -lmsvcrt -luser32 -lkernel32 -ladvapi32 -lshell32 "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtend.o"
set BCPLvn=%BCPLroot%\BCPL 2014-08-13\sial


for /f "usebackq tokens=*" %%i in (`getopt clOSo:x: %*`) do set args=%%i
for %%i in (%args%) do (
	if '!oarg!'=='?' (
	  set oarg=%%i
	)
	if '!farg!'=='?' (
	  set farg=%%i
	)
	if '!xasm!'=='?' (
	  set xasm=%%i
	)
	if '%%i'=='-O' (
	  set oflag=1
	)
	if '%%i'=='-o' (
	  set oarg=?
	)
	if '%%i'=='-c' (
	  set scode=1
	)
	if '%%i'=='-x' (
	  set xasm=?
	)
	if '%%i'=='-l' (
	  set obj=1
	)
	if '%%i'=='-S' (
	  set asm=1
	)
	if '%%i'=='--' (
	  set farg=?
	)
)
if '%farg%'=='?' goto usage
if '%oarg%'=='?' goto usage
if '%xasm%'=='?' goto usage

if defined xasm (
	echo.
	echo 	Cross-assembling for !xasm!
	echo.
) else (
	set xasm=386w
)

if not exist %farg% (
	echo "%0 %farg%: no such file"
	echo.
	goto usage
)
if not exist %farg% (
	echo "%0 %farg%: no such file"
	echo.
	goto usage
)
if not exist "%BCPLvn%\sial-%xasm%.exe" (
	echo 	sial-%xasm%: no such cross-assembler
	echo.
	goto usage
)

rem	if not exist libhdr cp "%BCPLocd%\src\libhdr" libhdr

if defined oarg (
    set f=%oarg%
) else (
	for /f "usebackq" %%i in (`basename %farg% .b`) do set f=%%i
)

if exist %f%.s del /q/f %f%.s > nul
if exist %f%.o del /q/f %f%.o > nul

"%BCPLvn%\bcpl2sial" %farg% %f%.sial
"%BCPLvn%\sial-%xasm%" %f%.sial %f%.s

if not defined scode if exist %f%.sial del /q/f %f%.sial > nul

if '%xasm%'=='386w' (

	as -o %f%.o %f%.s
rem	if not defined obj ld -Map ld.lst %LDFLAGS% -o %f%.exe %f%.o "%BCPLocd%\src\blib.o" "%BCPLocd%\src\global.o" "%BCPLocd%\src\su.o" "%BCPLocd%\src\rtc.o" %LDLIBS%
	if not defined obj ld %LDFLAGS% -o %f%.exe %f%.o "%BCPLocd%\src\blib.o" "%BCPLocd%\src\global.o" "%BCPLocd%\src\su.o" "%BCPLocd%\src\rtc.o" %LDLIBS%
	if not defined asm if exist %f%.s del /q/f %f%.s > nul
	if not defined obj if exist %f%.o del /q/f %f%.o > nul
) else (
	mv %f%.s %f%_%xasm%.s
)

:eof

set f=
set vn=
set rvn=
set edt=
set int=
set sial=
set version=
set oflag=
set asm=
set obj=
set oname=
set copt=
set args=
set LDFLAGS=
set LDLIBS=
