
SECTION "Subfil"

GET "libhdr"

STATIC {
  ch	= endstreamch
  item	= ?
  ext	= ?
                         // the following correspond to command-line switches:
  verbose	= FALSE				// if true talk to the user
  delsrce	= FALSE				// if true delete the source files
  listing	= FALSE				// if true list the contents of the .LIB file
  no_owrite	= FALSE				// if true we may NOT overwrite existing files
  }

STATIC {						// date and version number of Subfil
  vndate	= ?
  }

MANIFEST {
  vn	= 2						// version
  rvn	= 0						// revision
  edt	= 'c'					// edit
  }

/*
Change history:
	Vn	Rvn	Edt	Vn Date
	2	0	c	 2-Sep-2020		// addition of 'h' option and help text
	2	0	b	19-Mar-2018		// improved handling of newlines and comments in a '.sub' file
	2	0	a	18-Mar-2018		// added capability to compact files from a '.sub' file
	1	0	a	17-Mar-2018		// initial version, capable of extracting files from a '.lib' file
*/

MANIFEST {						// unusual ASCII character codes which appear in '.lib. files
  nul	= #x00
  ff	= #x0C
  }


LET exist(fname) = VALOF {		// check if 'filename' exists:  TRUE if so, FALSE if not
  LET strm, ipf = findinput(fname), input()
  TEST strm=0 THEN RESULTIS FALSE
  ELSE {
    selectinput(strm)
    endread()
	selectinput(ipf)
  }
  RESULTIS TRUE
}

AND rch() BE ch := rdch()		// ...because it's tiring to type 'ch := rdch()' everywhere

AND rditem() BE {				// read a word, white-space terminated, into 'item'
  LET v = VEC maxstrlength
  LET i = 1

  { v!i := ch
    rch()

    SWITCHON ch INTO {
	CASE ' ':
	CASE ',':
	CASE '*t':
	CASE '*n':
	CASE nul:
	CASE endstreamch:
      BREAK

    DEFAULT:
      i := i+1
      IF i >= 256 BREAK
    }
  } REPEAT

  v!0 := i
  packstring(v, item)
}

AND cmpstr(s1, s2) = VALOF {			// compare case-equated strings, returning 0 if so
  LET s1l, s2l = s1%0, s2%0
  IF s1l<s2l RESULTIS -1				// string1 shorter than string2
  IF s1l>s2l RESULTIS 1
  FOR i=1 TO s1l DO {
    LET c1, c2 = 'A' <= s1%i <= 'Z' -> s1%i-'A'+'a', s1%i, 'A' <= s2%i <= 'Z' -> s2%i-'A'+'a', s2%i
										// so we do a case-independent comparison
    IF c1 < c2 RESULTIS -1				// ith character in string1 has ASCII value less than ith character in string2
    IF c1 > c2 RESULTIS 1
  }
  RESULTIS 0							// strings are identical!
}

AND extn(fname) = VALOF {				// isolate any extension in filename and return it in 'ext', with length as result
  LET ptr, lth = ?, fname%0
  FOR i=1 TO lth DO IF fname%i='.' DO {
    ptr := i+1
    FOR j=ptr TO lth DO ext%(j-i) := fname%j
    BREAK
  }
  ext%0 := lth-ptr+1
  RESULTIS ext%0
}

AND swritef(strm, format, a, b, c, d, e, f, g, h, i, j, k) BE {			// execute 'writef()' to nominated stream
  LET opf = output()

  selectoutput(strm)
  writef(format, a, b, c, d, e, f, g, h, i, j, k)
  selectoutput(opf)
}



LET start(argv, argc) BE {
  LET fname, sourcestream, s, wrsub, lext = 0, ?, ?, TRUE, ?

  vndate := "2-Sep-2020"

  writef("subfil %s, %n.%n%c:*n*n", vndate, vn, rvn, edt)
  item := getvec(maxstrlength/bytesperword)
  ext := getvec(2)

  IF argc = 0 DO
    { writes("*nInvalid args to Subfil*n")
      abort(20)							// harsh, but what the hey...
    }

  FOR i=0 TO argc-1 DO {
    TEST (argv!i)%1 = '-' THEN {
	  FOR j=2 TO (argv!i)%0 DO {
        SWITCHON (argv!i)%j INTO {
          CASE 'c': CASE 'C':			// explode in 'chatty' mode. List file details as each file is written
          CASE 'v': CASE 'V':			// ...AKA 'verbose'
            verbose := TRUE
            ENDCASE
          CASE 'd': CASE 'D':			// delete the input file(s) on completion
            delsrce := TRUE
			wrsub := FALSE
            ENDCASE
          CASE 'h': CASE 'H':			// user guide...
            writef("subfil Vn: %n.%n%c, %s*n*n", vn, rvn, edt, vndate)
			writef("	subfil either:*n")
			writef("	concatenates a set of (text) files listed in a 'xxx.sub' file*n")
			writef("		'subfil xxx.sub', or*n")
			writef("	extracts the set of (text) files in a 'xxx.lib' file*n")
			writef("		'subfil xxx.lib'*n*n")
			writef("	subfil accepts the following options:*n")
			writef("		-c	or...*n")
			writef("		-v	- verbose mode;  list the files as they're processed*n")
			writef("		-d	- delete the input file(s) on completion*n")
			writef("		-h	- list this help text*n")
			writef("		-l	- list the contents of a 'xxx.lib' file*n")
			writef("		-n	- no overwrite of existing files*n*n")
            ENDCASE
          CASE 'l': CASE 'L':			// List the directory of the .LIB file
            listing := TRUE
			wrsub := FALSE
            ENDCASE
          CASE 'n': CASE 'N':			// Don't overwrite existing files.
            no_owrite := TRUE
            ENDCASE
          DEFAULT:
            writef("Switch '%c' is not implemented*n", (argv!i)%j)
        }
      }
    } ELSE {
      fname := argv!i
    }
  }
  IF fname=0 {
    writef("...no filename offered*n")
    stop(0)
  }

  lext := extn(fname)
  TEST cmpstr(ext, "sub") = 0 THEN {					// we're compacting...
    LET file, s, sin = getvec(maxstrlength/bytesperword), ?, ?

    sourcestream := findinput(fname)
    TEST sourcestream=0 THEN {
      writef("%s does not exist!*n", fname)
      stop(1)
    } ELSE {
      selectinput(sourcestream)
    }

    FOR i=1 TO fname%0 DO {
      file%i := fname%i
      IF file%i='.' BREAK
      file%0 := i
    }
    strcat(file, ".lib")

    IF no_owrite & exist(file) {
      writef("'%s' exists - will not overwrite!*n", file)
      stop(2)
    }

    IF verbose writef("Compacting from '%s' to '%s'*n", fname, file)

    s := findoutput(file)
	IF s {
      selectoutput(s)

      { rch()
        SWITCHON ch INTO {
          CASE '*n':
          CASE '*s':
          CASE '*t':
          CASE ff:
          CASE nul:
            LOOP
          CASE ';': {							// comment, apparently
              rch()
            } REPEATUNTIL ch='*n'				// ignore rest of line
            LOOP
          DEFAULT:
            rditem()									// read a fully qualified filename from the '.sub' file
        }
		IF cmpstr(item, "*n")=0 LOOP					// ignore blank lines
        IF ch=endstreamch BREAK
        sin := findinput(item)							// (attempt to) open the file named in '.sub'
        TEST sin THEN {									// file exists, add it to subfil library
          writef("SUBFILE: %s @%s %s <xxx> (sss)*n", item, timeofday(), date())		/*	write subfile header to '.lib' file
																						eg: "SUBFILE: 360LIB.GET @15:51:13 GMT Sun, 18 Mar 2018 <057> (128)"
																						nb: the PDP10 version of 'subfil' appeared to write the file's creation
																						date and time, file-access details and file-size as part of the header.
																						This version doesn't
																					*/
		  IF verbose swritef(sysout, "Compacting SUBFILE: %s @%s %s*n", item, timeofday(), date())
          selectinput(sin)
          { rch()
            IF ch=endstreamch BREAK
            wrch(ch)
          } REPEAT
          writes("\\\\\*n"); wrch(ff); writes("*n*n")	// write subfile trailer to '.lib' file
          endread()										// close the subfile...
          IF delsrce delete(item)						// ...delete it if required...
          selectinput(sourcestream)						// ...and reselect the '.sub' command file
        } ELSE {
          swritef(syserr, "file '%s' named in '%s' fails to open*n", item, fname)
        }
      } REPEAT
    }
    endwrite()

  } ELSE {
  TEST cmpstr(ext, "lib") = 0 {							// we're extracting...
    IF verbose writef("Expanding '%s'*n", fname)
    sourcestream := findinput(fname)
    TEST sourcestream=0 THEN {
      writef("%s does not exist!*n", fname)
      stop(3)
    } ELSE {
      selectinput(sourcestream)
    }

    IF listing writef("Subfiles of '%s':*n", fname)
    { rch()
      SWITCHON ch INTO {
        CASE 'S':
          rditem()
          IF cmpstr(item, "SUBFILE:")=0 THEN {			// found a subfile header, eg: "SUBFILE: 360LIB.BCL @16:50 9-JUN-1978 <057> (1408)" line
            rch()										// dispose of separating character
            rditem()									// ...next item will be the filename

            IF no_owrite & exist(item) {
              writef("'%s' exists - will not overwrite!*n", item)
              wrsub := FALSE
            }

            IF verbose writes("Extracting ")
            IF verbose | listing writef("SUBFILE: '%s'" , item)
            UNTIL ch='*n' | ch=endstreamch DO {
              IF (verbose | listing) DO wrch(ch)		// display or discard rest of line
              rch()
            }
            IF verbose | listing wrch('*n')
            IF wrsub s := findoutput(item)
            TEST s | NOT wrsub THEN {
              IF wrsub selectoutput(s)
              { rch()
                SWITCHON ch INTO {
                  CASE endstreamch:
		            IF wrsub endwrite()					// marker for end of subfile
                    BREAK
                    ENDCASE

                  CASE nul:								// NUL
                  CASE ff:								// FF
                    LOOP								// don't write to output!

                  CASE '\':
                    rditem()
		            IF cmpstr(item, "\\\\\")=0 THEN {
		              IF wrsub endwrite()				// marker for end of subfile
                      IF NOT listing wrsub := TRUE		// ...so we may potentially write further subfiles if they don't already exist
                      BREAK
                    }
                    IF wrsub writes(item)

                  DEFAULT:
                    IF wrsub wrch(ch)
                }
              } REPEAT
            } ELSE writef("Problem writing to *"%s*"*n", item)
          }
          ENDCASE

        DEFAULT:
        }
      } REPEATUNTIL ch=endstreamch
      endread()											// close the library file...
      IF delsrce delete(fname)							// ...and delete it if required
    } ELSE {
      writef("File extension '.%s' not recognised*n", ext)
    }
  }
}
