@echo off

rem	Copyright (c) 2021 Dave Cannon.  All rights reserved.
rem	$Id: bcpl.cmd,v 2.00 10-Mar-2021 11:35:02 rn Exp $

setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

set vn=2
set rvn=00
set edt=10-Mar-2021

set version=%0 Vn %vn%.%rvn%	%edt%

rem	Vn 2.00	10-Mar-2021	Amended for INTCODE-style BCPL libraries and run-time
rem	Vn 1.00	19-Aug-2014	Initial version for DOS under Windows (amended from the bcpl-sial version)

if not '%1'=='' goto doit

:usage
echo	Usage:	%0 [-Options] [-o output] file
echo		where	file is a BCPL source program to be compiled
echo			[-o output] (optional) specifies the compiled program name
echo			[-Options] may be any of:
echo			[-O] (optional) suppresses the code optimiser
echo			[-c] (optional) preserve MCODE output as 'file.mcode'
echo			[-l] (optional) preserve the 'file.o' object file for ld
echo			[-S] (optional) preserve the 'file.s' assembler file
echo.
echo		%version%
echo.
goto eof

:doit

set oflag=0
set args=

if not defined GnuWin (
  echo.
  echo 	The BCPL environment variables have not been established.
  echo 	You must execute ..\BCPLenv.cmd before running this script
  echo.
  goto eof
)

set LDFLAGS=-Bdynamic -T"%GnuWin%\mingw32\lib\ldscripts\i386pe_mgw.x" "%GnuWin%\lib\crt2.o" "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtbegin.o"
set LDLIBS=-lmingw32 -lmoldname -lmingwex -lmsvcrt -luser32 -lkernel32 -ladvapi32 -lshell32 "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtend.o"
set MCPLlib=%BCPLroot%\BCPL 2020-05-13\mcpl\MCPLlib
set MCPLvn=%BCPLroot%\BCPL 2020-05-13\mcpl


for /f "usebackq tokens=*" %%i in (`getopt clOSo: %*`) do set args=%%i
for %%i in (%args%) do (
	if '!oarg!'=='?' (
	  set oarg=%%i
	)
	if '!farg!'=='?' (
	  set farg=%%i
	)
	if '%%i'=='-O' (
	  set oflag=1
	)
	if '%%i'=='-o' (
	  set oarg=?
	)
	if '%%i'=='-c' (
	  set mcode=1
	)
	if '%%i'=='-l' (
	  set obj=1
	)
	if '%%i'=='-S' (
	  set asm=1
	)
	if '%%i'=='--' (
	  set farg=?
	)
)
if '%farg%'=='?' goto usage
if '%oarg%'=='?' goto usage

if not exist %farg% (
	echo "%0 %farg%: no such file"
	echo.
	goto usage
)

rem	if not exist libhdr cp "%BCPLsrc%\libhdr" libhdr

if defined oarg (
    set f=%oarg%
) else (
	for /f "usebackq" %%i in (`basename %farg% .m`) do set f=%%i
)

if exist %f%.s del /q/f %f%.s > nul
if exist %f%.o del /q/f %f%.o > nul

"%MCPLvn%\mcpl2mial" %farg% %f%.mial
"%MCPLvn%\mial-386_w" %f%.mial %f%.s

if not defined mcode if exist %f%.mial del /q/f %f%.mial > nul
if defined mcode (
  if exist %f%.mcode del /q/f %f%.mcode
  if exist MCODE ren MCODE %f%.mcode
) else (
  if exist MCODE del /q/f MCODE > nul
)

as -o %f%.o %f%.s
if not defined obj ld %LDFLAGS% -o %f%.exe %f%.o "%MCPLlib%\icmlib.o" "%MCPLlib%\global.o" "%BCPLsrc%\su.o" "%BCPLsrc%\rtc.o" %LDLIBS%

if not defined asm if exist %f%.s del /q/f %f%.s > nul
if not defined obj if exist %f%.o del /q/f %f%.o > nul

:eof

set d=
set f=
set vn=
set rvn=
set edt=
set int=
set sial=
set version=
set gnuwin=
set oflag=
set asm=
set obj=
set oname=
set copt=
set args=
set LDFLAGS=
set LDLIBS=
