@echo off

rem	$Id: bcpl.cmd,v 2.09 8-Mar-2021 08:30:21 rn Exp $
rem	Copyright (c) 2013-2020 Dave Cannon.  All rights reserved.

setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

set vn=2
set rvn=09
set edt=8-Mar-2021

set version=%0 Vn %vn%.%rvn%	%edt%

rem	Vn 2.09	 8-Mar-2021 run the 1979 sial test compiler
rem	Vn 2.08	 1-Nov-2020 add code to pass named run-time parameters to compiler
rem	Vn 2.07	23-Mar-2020 compile with BCPL version with EXTERNAL and in-line CODE support
rem	Vn 2.06	11-Jan-2016 change order of concatenation of INTCODE files so 'xg_i'
rem						generates appropriate/helpful diagnostics
rem	Vn 2.05	20-Aug-2014 all environment variables now set externally by ..\BCPLenv.cmd
rem	Vn 2.04	24-Aug-2013 introduction of 'getopt' to parse command line options
rem	Vn 2.03	 9-Feb-2013 BCPLHDRS environment variable added as location of 'libhdr', etc.
rem	Vn 2.02	 5-Feb-2013 adapted to run the 1-Feb-2011 version of the BCPL compiler
rem	Vn 2.01	 7-Jan-2013 Employs xg Vn 2.03 developed to translate INTCODE to assembler
rem	Vn 2.00	 6-Jan-2013	modified to run Martin Richards' latest compiler version

rem	Vn 1.05	26-Oct-2012	add '-c' option to preserve the OCODE output as'file.ocd'
rem	Vn 1.04	12-Oct-2012	add '-i' option to preserve the INTCODE output as'file.i'
rem	Vn 1.03	09-Oct-2012	add '-l' option to preserve the 'file.o' output for ld
rem	Vn 1.02	29-Sep-2012	add '-S' option to preserve the 'file.asm' output
rem	Vn 1.01	28-Sep-2012	add quotes to variable 'd' for compatibility with Windows XP
rem	Vn 1.00	27-Sep-2012	Initial translation from Robert Nordier's shell script to DOS

if not '%1'=='' goto doit

:usage
echo	Usage:	%0 [-Options] [-o output] file
echo		where	file is a BCPL source program to be compiled
echo			[-o output] (optional) specifies the compiled program name
echo			[-Options] may be any of:
echo			[-O] (optional) invokes the copt code optimiser
echo			[-c] (optional) preserve OCODE output as 'file.ocd'
echo			[-i] (optional) preserve INTCODE output as 'file.i'
echo			[-l] (optional) preserve the 'file.o' output for ld
echo			[-S] (optional) preserve the 'file.asm' output
echo.
echo		%version%
echo.
goto end


:doit

set oflag=0
set args=

if not defined GnuWin (
  echo.
  echo 	The BCPL environment variables have not been established.
  echo 	You must execute ..\BCPLenv.cmd before running this script
  echo.
  goto end
)

set LDFLAGS=-Bdynamic -T"%GnuWin%\mingw32\lib\ldscripts\i386pe_mgw.x" "%GnuWin%\lib\crt2.o" "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtbegin.o"
set LDLIBS=-lmingw32 -lmoldname -lmingwex -lmsvcrt -luser32 -lkernel32 -ladvapi32 -lshell32 "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtend.o"
set BCPLvn=%BCPLroot%\BCPL 2020-05-13\sial\BCPL 1979-01-01

rem	Compiler parameters may be offered by name, preceded by a '/' character and space-separated, after
rem	the filename to be compiled, eg:

	rem			bcpl -i xref.b /tree /eqcases /size:50000

	rem			NB: some options expect a string or value to be offered, as with 'size:50000' above


	rem				Position	Param
	rem				--------	-----
	rem				(0) 	and...
	rem				(1) 	...are reserved for input and output filenames - this script already provides these!
set s2=0&	rem		(2)			ver:<filename>
set s3=0&	rem		(3)			size:<treesize>
set s4=0&	rem		(4)			tree
set s5=0&	rem		(5)			nonames
set s6=0&	rem		(6)			d1
set s7=0&	rem		(7)			d2
set s8=0&	rem		(8)			oender
set s9=0&	rem		(9)			eqcases
set s10=0&	rem		(10)		bin
set s11=0&	rem		(11)		xref
set s12=0&	rem		(12)		gdefs
set s13=0&	rem		(13)		hdrs:<envname>
set s14=0&	rem		(14)		gb2312
set s15=0&	rem		(15)		utf8
set s16=0&	rem		(16)		savesize:<savespacesize>	(change this at your peril!)
set s17=0&	rem		(17)		hard
set s18=0&	rem		(18)		t32
set s19=0&	rem		(19)		t64
set s20=0&	rem		(20)		opt:<optstring>


for /f "usebackq tokens=*" %%i in (`getopt cilOSo: %*`) do set args=%%i
rem	echo 	args='%args%'
for %%i in (%args%) do (
	if '!oarg!'=='?' (
	  set oarg=%%i
	)
	if '!farg!'=='?' (
	  set farg=%%i
	)
	if '%%i'=='-O' (
	  set oflag=1
	)
	if '%%i'=='-o' (
	  set oarg=?
	)
	if '%%i'=='-c' (
	  set ocd=1
	)
	if '%%i'=='-i' (
	  set int=1
	)
	if '%%i'=='-l' (
	  set obj=1
	)
	if '%%i'=='-S' (
	  set asm=1
	)
	if '%%i'=='--' (
	  set farg=?
	)
	set op=%%i
	set op=!op:~0,1!
	if '!op!'=='/' (
	  set op=%%i
	  set op=!op:~1!
	  for /f "tokens=1,2 delims=:" %%i in ("!op!") do (
		set op=%%i
		set val=%%j
	  )
rem	echo op=!op!, val='!val!'
	  if '!op!' == 'ver'		set s2=!val!
	  if '!op!' == 'size'		set s3=!val!
	  if '!op!' == 'tree'		set s4=1
	  if '!op!' == 'nonames'	set s5=1
	  if '!op!' == 'd1'			set s6=1
	  if '!op!' == 'd2'			set s7=1
	  if '!op!' == 'oender'		set s8=1
	  if '!op!' == 'eqcases'	set s9=1
	  if '!op!' == 'bin'		set s10=1
	  if '!op!' == 'xref'		set s11=1
	  if '!op!' == 'gdefs'		set s12=1
	  if '!op!' == 'hdrs'		set s13=!val!
	  if '!op!' == 'gb2312'		set s14=1
	  if '!op!' == 'utf8'		set s15=1
	  if '!op!' == 'savesize'	set s16=!val!
	  if '!op!' == 'hard'		set s17=1
	  if '!op!' == 't32'		set s18=1
	  if '!op!' == 't64'		set s19=1
	  if '!op!' == 'opt'		set s20=!val!
rem	  echo !op!: s9=!s9!
	)
)

:loop
  set /a x+=1
  if !x! gtr 20 goto fi
  set opts=!opts! !s%x%!
  goto loop
:fi
rem	echo opts='%opts%'

if '%farg%'=='?' goto usage
if '%oarg%'=='?' goto usage

if not exist %farg% (
	echo "%0 %farg%: no such file"
	goto usage
)

if exist OCODE del /q/f OCODE > nul
if exist INTCODE del /q/f INTCODE > nul
if exist ASM del /q/f ASM > nul

rem if not exist libhdr cp %BCPLsrc%\libhdr+ libhdr

if defined oarg (
    set f=%oarg%
) else (
	for /f "usebackq" %%i in (`basename %farg% .b`) do set f=%%i
)

if exist %f%.ocd del /q/f %f%.ocd > nul
if exist %f%.int del /q/f %f%.int > nul
if exist %f%.asm del /q/f %f%.asm > nul
if exist %f%.s del /q/f %f%.s > nul
if exist %f%.o del /q/f %f%.o > nul

IF defined ocd (
	"%BCPLvn%\bcpl+" %farg% 0 %opts%
	move OCODE %f%.ocd > nul
) else (
	"%BCPLvn%\bcpl+" %farg% %f%.i %opts%
rem	cat %f%.i "%BCPLlib%\blib.i" "%BCPLvn%\codegens\iclib.i" > %f%.int
	cat %f%.i "%BCPLlib%\blib.i" "%BCPLlib%\iclib.i" > %f%.int
	if exist OCODE del /q/f OCODE > nul
	if defined int (
		move %f%.i %f%.int > nul
		goto end
	) else (
		del /q/f %f%.i > nul
	)
	"%BCPLvn%\..\..\codegens\xg_i" < %f%.int
	if %oflag% equ 1 (
		"%BCPLcopt%\copt" "%BCPLsrc%\rules" < ASM > ASM.op && mv ASM.op ASM
	)
	move ASM %f%.s > nul
	as -o %f%.o %f%.s
	if not defined obj if exist %f%.o ld %LDFLAGS% -o %f%.exe %f%.o "%BCPLsrc%\su.o" "%BCPLsrc%\rtc.o" %LDLIBS%
)

if exist %f%.int del /q/f %f%.int > nul
if defined asm (
	if exist %f%.s move %f%.s %f%.asm > nul
	)
if exist %f%.s del /q/f %f%.s > nul
if not defined obj if exist %f%.o del /q/f %f%.o > nul

:end

set f=
set vn=
set rvn=
set edt=
set int=
set ocd=
set version=
set oflag=
set asm=
set obj=
set oarg=
set args=
set LDFLAGS=
set LDLIBS=

:eof
