SECTION "sial-ocode"

//	Convert a SIAL source file to human-readable OCODE form...

GET "libhdr"

GET "sial.h"

GLOBAL {
sialin: ug
sialout; stdin; stdout

rdf; rdp; rdg; rdk; rdh; rdw; rdl; rdd; rdc
rdcode

scan
cvf; cvfp; cvfg; cvfk; cvfh; cvfw; cvfl; cvfd
}

MANIFEST {
savespacesize = 2
}

STATIC {
ssp = 0
}

LET start() = VALOF
{ LET argv = VEC 20

   sialout := 0
   stdout := output()
   IF rdargs("FROM,TO/K", argv, 20)=0 DO
   { writes("Bad args for sial-ocode*n")
      RESULTIS 20
   }
   IF argv!0=0 DO argv!0 := "prog.sial"
   IF argv!1=0 DO argv!1 := "prog.ocode"
   sialin := findinput(argv!0)
   IF sialin=0 DO
   { writef("Trouble with file %s*n", argv!0)
      RESULTIS 20
   }
   sialout := findoutput(argv!1)
   
   IF sialout=0 DO
   { writef("Trouble with file %s*n", argv!1)
      RESULTIS 20
   }
   
   writef("Converting %s to %s*n", argv!0, argv!1)
   selectinput(sialin)
   selectoutput(sialout)
   scan()
   endread()
   UNLESS sialout=stdout DO endwrite()
   selectoutput(stdout)
   writef("Conversion complete*n")
   RESULTIS 0
}

// argument may be of form Ln
AND rdcode(let) = VALOF
{ LET a, ch, neg = 0, ?, FALSE

   ch := rdch() REPEATWHILE ch='*s' | ch='*n'

   IF ch=endstreamch RESULTIS -1

   UNLESS ch=let DO error("Bad item, looking for %c found %c*n", let, ch)

   ch := rdch()

   IF ch='-' DO { neg := TRUE; ch := rdch() }

   WHILE '0'<=ch<='9' DO { a := 10*a + ch - '0'; ch := rdch()  }

   RESULTIS neg -> -a, a
}

AND rdf() = rdcode('F')
AND rdp() = rdcode('P')
AND rdg() = rdcode('G')
AND rdk() = rdcode('K')
AND rdh() = rdcode('H')
AND rdw() = rdcode('W')
AND rdl() = rdcode('L')
AND rdm() = rdcode('M')
AND rdc() = rdcode('C')

AND error(mess, a, b, c) BE
{ LET out = output()
   UNLESS out=stdout DO
   { selectoutput(stdout)
      writef(mess, a, b, c)
      selectoutput(out)
   }
   writef(mess, a, b, c)
}

AND scan() BE
{ LET op = rdf()

   SWITCHON op INTO

   { DEFAULT:       error("Bad op %n*n", op); LOOP

      CASE -1:       RETURN
      
      CASE f_lp:     cvfp("LP", op); ENDCASE
      CASE f_lg:     cvfg("LG", op); ENDCASE
      CASE f_ll:     cvfl("LL"); ENDCASE

      CASE f_llp:    cvfp("LLP", op); ENDCASE
      CASE f_llg:    cvfg("LLG", op); ENDCASE
      CASE f_lll:    cvfl("LLL", op); ENDCASE
      CASE f_lf:     cvfl("LF", op); ENDCASE
      CASE f_lw:     cvfm("LW"); ENDCASE

      CASE f_l:      cvfk("LN", op); ENDCASE
      CASE f_lm:     cvfk("LM", op); ENDCASE

      CASE f_sp:     cvfp("SP", op); ENDCASE
      CASE f_sg:     cvfg("SG", op); ENDCASE
      CASE f_sl:     cvfl("SL", op); ENDCASE

      CASE f_ap:     cvfp("AP", op); ENDCASE
      CASE f_ag:     cvfg("AG", op); ENDCASE
      CASE f_a:      cvfk("A", op); ENDCASE
      CASE f_s:      cvfk("S", op); ENDCASE

      CASE f_lkp:    cvfkp("LKP", op); ENDCASE
      CASE f_lkg:    cvfkg("LKG", op); ENDCASE
      CASE f_rv:     cvf("RV", op); ENDCASE
      CASE f_rvp:    cvfp("RVP", op); ENDCASE
      CASE f_rvk:    cvfk("RVK", op); ENDCASE
      CASE f_st:     cvf("STIND", op); ENDCASE
      CASE f_stp:    cvfp("STP", op); ENDCASE
      CASE f_stk:    cvfk("STK", op); ENDCASE
      CASE f_stkp:   cvfkp("STKP", op); ENDCASE
      CASE f_skg:    cvfkg("SKG", op); ENDCASE
      CASE f_xst:    cvf("XST", op); ENDCASE

      CASE f_k:      cvfp("K", op); ENDCASE
      CASE f_kpg:    cvfpg("KPG"); ENDCASE

      CASE f_neg:    cvf("NEG", op); ENDCASE
      CASE f_not:    cvf("NOT", op); ENDCASE
      CASE f_abs:    cvf("ABS", op); ENDCASE

      CASE f_xdiv:   cvf("XDIV", op); ENDCASE
      CASE f_xrem:   cvf("XREM", op); ENDCASE
      CASE f_xsub:   cvf("XSUB", op); ENDCASE

      CASE f_mul:    cvf("MUL", op); ENDCASE
      CASE f_div:    cvf("DIV", op); ENDCASE
      CASE f_rem:    cvf("REM", op); ENDCASE
      CASE f_add:    cvf("ADD", op); ENDCASE
      CASE f_sub:    cvf("SUB", op); ENDCASE

      CASE f_eq:     cvf("EQ", op); ENDCASE
      CASE f_ne:     cvf("NE", op); ENDCASE
      CASE f_ls:     cvf("LS", op); ENDCASE
      CASE f_gr:     cvf("GR", op); ENDCASE
      CASE f_le:     cvf("LE", op); ENDCASE
      CASE f_ge:     cvf("GE", op); ENDCASE
      CASE f_eq0:    cvf("EQ0", op); ENDCASE
      CASE f_ne0:    cvf("NE0", op); ENDCASE
      CASE f_ls0:    cvf("LS0", op); ENDCASE
      CASE f_gr0:    cvf("GR0", op); ENDCASE
      CASE f_le0:    cvf("LE0", op); ENDCASE
      CASE f_ge0:    cvf("GE0", op); ENDCASE

      CASE f_lsh:    cvf("LSHIFT", op); ENDCASE
      CASE f_rsh:    cvf("RSHIFT", op); ENDCASE
      CASE f_and:    cvf("LOGAND", op); ENDCASE
      CASE f_or:     cvf("LOGOR", op); ENDCASE
      CASE f_xor:    cvf("XOR", op); ENDCASE
      CASE f_eqv:    cvf("EQV", op); ENDCASE

      CASE f_gbyt:   cvf("GETBYTE", op);  ENDCASE
      CASE f_xgbyt:  cvf("GETBYTE", op); ENDCASE
      CASE f_pbyt:   cvf("PUTBYTE", op);  ENDCASE
      CASE f_xpbyt:  cvf("XPBYT", op); ENDCASE

      CASE f_swb:    cvswb(); ENDCASE
      CASE f_swl:    cvswl(); ENDCASE

      CASE f_xch:    cvf("XCH", op); ENDCASE
      CASE f_atb:    LOOP; 						// cvf("ATB", op); ENDCASE
      CASE f_atc:    LOOP;						// cvf("ATC", op); ENDCASE
      CASE f_bta:    LOOP;						// cvf("BTA", op); ENDCASE
      CASE f_btc:    cvf("BTC", op); ENDCASE
      CASE f_atblp:  cvfp("ATBLP", op); ENDCASE
      CASE f_atblg:  cvfg("ATBLG", op); ENDCASE
      CASE f_atbl:   cvfk("ATBL", op); ENDCASE

      CASE f_j:      cvfl("JUMP", op); ENDCASE
      CASE f_rtn:    cvf("RTRN", op); ENDCASE
      CASE f_goto:   cvf("GOTO", op); ENDCASE

      CASE f_ikp:    cvfkp("IKP", op); ENDCASE
      CASE f_ikg:    cvfkg("IKG", op); ENDCASE
      CASE f_ikl:    cvfkl("IKL"); ENDCASE
      CASE f_ip:     cvfp("IP", op);   ENDCASE
      CASE f_ig:     cvfg("IG", op);   ENDCASE
      CASE f_il:     cvfl("IL", op);   ENDCASE

      CASE f_jeq:    cvfl("JEQ", op); ENDCASE
      CASE f_jne:    cvfl("JNE", op); ENDCASE
      CASE f_jls:    cvfl("JLS", op); ENDCASE
      CASE f_jgr:    cvfl("JGR", op); ENDCASE
      CASE f_jle:    cvfl("JLE", op); ENDCASE
      CASE f_jge:    cvfl("JGE", op); ENDCASE
      CASE f_jeq0:   cvfl("JEQ0", op); ENDCASE
      CASE f_jne0:   cvfl("JNE0", op); ENDCASE
      CASE f_jls0:   cvfl("JLS0", op); ENDCASE
      CASE f_jgr0:   cvfl("JGR0", op); ENDCASE
      CASE f_jle0:   cvfl("JLE0", op); ENDCASE
      CASE f_jge0:   cvfl("JGE0", op); ENDCASE
      CASE f_jge0m:  cvfm("JGE0", op); ENDCASE

      CASE f_brk:    cvf("BRK", op); ENDCASE
      CASE f_nop:    cvf("NOP", op); ENDCASE
      CASE f_chgco:  cvf("CHGCO", op); ENDCASE
      CASE f_mdiv:   cvf("MDIV", op); ENDCASE
      CASE f_sys:    cvf("SYS", op); ENDCASE

      CASE f_section:  cvfs("SECTION"); ENDCASE
      CASE f_modstart: cvf("MODSTART", op); ENDCASE
      CASE f_modend:   cvf("MODEND", op); ENDCASE
      CASE f_global:   cvglobal(); ENDCASE
      CASE f_string:   cvstring(); ENDCASE
      CASE f_const:    cvconst(); ENDCASE
      CASE f_static:   cvstatic(); ENDCASE
      CASE f_mlab:     cvfm("MLAB"); ENDCASE
      CASE f_lab:      cvfl("LAB", op); ENDCASE
      CASE f_lstr:     cvfm("LSTR"); ENDCASE
      CASE f_entry:    cventry(); ENDCASE

      CASE f_float:    cvf("FLOAT", op); ENDCASE
      CASE f_fix:      cvf("FIX", op); ENDCASE
      CASE f_fabs:     cvf("FABS", op); ENDCASE

      CASE f_fmul:     cvf("FMUL", op); ENDCASE
      CASE f_fdiv:     cvf("FDIV", op); ENDCASE
      CASE f_fadd:     cvf("FADD", op); ENDCASE
      CASE f_fsub:     cvf("FSUB", op); ENDCASE
      CASE f_fneg:     cvf("FNEG", op); ENDCASE

      CASE f_feq:      cvf("FEQ", op); ENDCASE
      CASE f_fne:      cvf("FNE", op); ENDCASE
      CASE f_fls:      cvf("FLS", op); ENDCASE
      CASE f_fgr:      cvf("FGR", op); ENDCASE
      CASE f_fle:      cvf("FLE", op); ENDCASE
      CASE f_fge:      cvf("FGE", op); ENDCASE
   }

   newline()
} REPEAT

AND cvf(s, op)  BE
{  SWITCHON op INTO {
     CASE f_xst:
       writef("	STIND");	ENDCASE
     DEFAULT:
       writef("	%s", s)
   }
}

AND cvfp(s, op) BE
{  LET P = rdp()
   AND shwm = savespacesize+P
   
   SWITCHON op INTO {
     CASE f_k:
       writef("	STACK %n*n", shwm)
       writef("	RTAP %n", P)
       ssp := shwm;			ENDCASE

     CASE f_stp:
     CASE f_rvp:
     CASE f_ap:
       writef("	LP %n*n", P)
       writef("	ADD")
       IF op=f_stp {
         newline()
         writef("	STIND")
       }
       IF op=f_rvp {
         newline()
         writef("	RV")
       };					ENDCASE
     CASE f_atblp:
       s := "LP"
     CASE f_sp:
       ssp := P
     DEFAULT:
       writef("	%s %n", s, P)
   }
}
AND cvfkp(s, op) BE
{  LET K, P = rdk(), rdp()
   SWITCHON op INTO {
     CASE f_ikp:
       writef("	LP %n*n", P)
       writef("	LN %n*n", K)
       writef("	ADD*n")
       writef("	SP %n", P);	ENDCASE
     CASE f_stkp:
     CASE f_lkp:
       writef("	LP %n*n", P)
       IF K>0 {
         writef("	LN %n*n", K)
         writef("	ADD*n")
       }
       TEST op=f_stkp THEN
         writef("	STIND")
       ELSE
         writef("	RV");	ENDCASE
     DEFAULT:
       writef("	%s K%n P%n", s, K, P)
   }
}
AND cvfg(s, op) BE
{  LET G = rdg()
   SWITCHON op INTO {
     CASE f_lg:
     CASE f_atblg:
       writef("	%s %n", "LG", G);	ENDCASE
     CASE f_ag:
       writef("	%s %n*n", "LG", G)
       writef("	ADD");		ENDCASE
     DEFAULT:
       writef("	%s %n", s, G)
   }
//writef("	%s %n", s, rdg())
}
AND cvfkg(s, op) BE
{  LET K, G = rdk(), rdg()
   writef("	LG %n*n", G)
   writef("	LN %n*n", K)
   writef("	ADD*n")

   SWITCHON op INTO {
     CASE f_lkg:
       writef("	RV");	ENDCASE
     CASE f_ikg:
       writef("	SG %n", G);	ENDCASE
     CASE f_skg:
       writef("	STIND");	ENDCASE
     DEFAULT:
       writef("	%s K%n G%n", s, K, G)
   }
}
AND cvfkl(s) BE writef("	%s K%n L%n", s, rdk(), rdl())

AND cvfpg(s) BE
{  LET P, G = rdp(), rdg()
   AND shwm = savespacesize+P
 //shwm := ssp>shwm -> ssp, shwm
   writef("	STACK %n*n", shwm)
   writef("	LG %n*n", G)
   writef("	RTAP %n", P)
   ssp := shwm
 //writef("	%s P%n G%n", s, rdp(), rdg())
}
AND cvfk(s, op) BE
{  LET K = rdk()
   SWITCHON op INTO {
     CASE f_stk:
       writef("	LN %n*n", K)
       writef("	ADD*n")
	   writef("	STIND");	ENDCASE
     CASE f_s:
       writef("	LN %n*n", K)
       writef("	SUB");		ENDCASE
     CASE f_a:
       writef("	LN %n*n", K)
       writef("	ADD");		ENDCASE
     CASE f_lm:
       K := -K
     CASE f_atbl:
       s := "LN"
     DEFAULT:
       writef("	%s %n", s, K)
   }
}

AND cvfh(s) BE writef("	%s H%n", s, rdh())

AND cvfw(s) BE writef("	%s W%n", s, rdw())

AND cvfl(s, op) BE
{  LET L = rdl()
   SWITCHON op INTO {
     CASE f_jeq0:
  //   writef("	LN 0*n")
       writef("	JF L%n", L);	ENDCASE
     CASE f_jeq:
       writef("	EQ*n")
       writef("	JT L%n", L);	ENDCASE
     CASE f_jne0:
  //   writef("	LN 0*n")
       writef("	JF L%n", L);	ENDCASE
     CASE f_jne:
       writef("	NE*n")
       writef("	JT L%n", L);	ENDCASE
    CASE f_jge0:
    CASE f_jge0m:
       writef("	LN 0*n")
     CASE f_jge:
       writef("	GE*n")
       writef("	JT L%n", L);	ENDCASE
     CASE f_jgr:
       writef("	GR*n")
       writef("	JT L%n", L);	ENDCASE
     CASE f_jle0:
       writef("	LN 0*n")
     CASE f_jle:
       writef("	LE*n")
       writef("	JT L%n", L);	ENDCASE
     CASE f_jls:
       writef("	LS*n")
       writef("	JT L%n", L);	ENDCASE
     CASE f_lf:
       writef("	LF L%n", L);	ENDCASE
     DEFAULT:
       writef("%s%s L%n", s%1='J' -> "	", "", s, L)
   }
}

AND cvfm(s) BE writef("	%s M%n", s, rdm())

AND cvswl() BE
{  LET n = rdk()
   LET l = rdl()
   writef("SWL K%n L%n", n, l)
   FOR i = 1 TO n DO writef("*n	L%n", rdl())
}

AND cvswb() BE
{  LET n = rdk()
   LET l = rdl()
   writef("SWB K%n L%n", n, l)
   FOR i = 1 TO n DO 
   { LET k = rdk()
      LET l = rdl()
      writef("*nK%n L%n", k, l)
   }
}

AND cvglobal() BE
{  LET n = rdk()
   writef("GLOBAL %n*n", n)
   FOR i = 1 TO n DO
   { LET g = rdg()
      LET n = rdl()
      writef("	%n	L%n*n", g, n)
   }
   writef("	%n", rdg())
}

AND cvstring() BE
{  LET lab = rdm()
   LET n = rdk()
   writef("LAB M%n %n", lab, n)
   FOR i = 1 TO n DO {
     writef(" '%c'", rdc())
	 IF (i MOD 14) = 0 newline()
   }
 //FOR i = 1 TO n DO wrch(rdc()); wrch('"')
}

AND cvconst() BE
{  LET lab = rdm()
   LET w = rdw()
   writef("CONST   M%n W%n", lab, w)
}

AND cvstatic() BE
{  LET lab = rdl()
   LET n = rdk()
   writef("DATALAB	L%n*n", lab)
   FOR i = 1 TO n DO writef("	ITEMN %n*n", rdw())
}

AND cvfs(s) BE
{  LET n = rdk()
   writef("%s %n *"", s, n)
 //FOR i = 1 TO n DO writef(" C%n", rdc())
   FOR i = 1 TO n DO wrch(rdc()); wrch('"')
}

AND cventry() BE
{  LET n = rdk()
   LET v = VEC 256
   v%0 := n
   FOR i = 1 TO n DO v%i := rdc()
   writef("*n// Entry to: %s*n", v)
   writef("ENTRY %n *"%s*"*n", n, v)
   writef("	SAVE 2")
 //FOR i = 1 TO n DO writef(" C%n", v%i)
 //writes(v); wrch('"')
}
