@echo off

rem	Establish environment variables for BCPL scripts

set vn=1
set rvn=10
set edt=12-Apr-2020

rem	Vn 1.10	12-Apr-2020	Recoded section to determine 'minGWVn'

rem	Vn 1.04	 4-Jan-2016	Recoded fix to finding home folder for 'GnuWin'
rem	Vn 1.03	 3-Jan-2016	Attempted fix to finding home folder for 'GnuWin'
rem	Vn 1.02	20-Oct-2015	MCPLhdrs added
rem	Vn 1.01	15-Sep-2014	Execute from source directory
rem	Vn 1.00	20-Aug-2014	Initial version

if '%1'=='-l' goto listenv
if '%1'=='/?' goto help
if '%1'=='-h' (
  :help
  echo	Usage:	%0 [/? ^| -h ^| -l ^| -d ^| -q]
  echo		where	%0 ^(no options^) sets and lists the BCPL environment
  echo			[-q] sets the BCPL environment - no output
  echo			[/? ^| -h] lists this help text
  echo			[-l] lists the BCPL environment, if set
  echo			[-d] deletes the BCPL environment
  echo.
  echo			The BCPL environment variables are:
  echo.
  echo			BCPLenv:  the version no: ^(%vn%.%rvn% %edt%^) of this script
  echo			BCPLcopt: your copt ^(code optimiser^) installation directory
  echo			BCPLhdrs: your BCPL header ^(GET file^) installation directory
  echo			BCPLlib:  your BCPL library ^(loader object file^) installation
  echo				  directory
  echo			BCPLocd:  your BCPL Ocode compiler installation directory,
  echo				  if installed
  echo			BCPLroot: the root of the BCPLkit family of compilers
  echo			BCPLsrc:  your BCPL compiler installation directory
  echo			MCPLhdrs: your MCPL header ^(GET file^) installation directory
  echo			minGWVn:  your MinGW installation version number
  echo			GnuWin:   your MinGW ^(GNU for Windows^) c compiler/assembler
  echo				  /linker installation directory
  echo.
  pause
  :listenv
  if defined BCPLenv (
    goto envset
  ) else (
    echo.
	echo 	BCPL environment has not been established
	echo.
	goto eof
  )
)

rem Initialize/purge BCPL environment

set m_vn=
set BCPLcopt=
set BCPLdir=
set BCPLenv=
set BCPLhdrs=
set BCPLlib=
set BCPLroot=
set BCPLsrc=
set BCPLocd=
set GnuWin=
set MCPLhdrs=
set minGWVn=

if '%1'=='-d' (
  echo 	BCPL environment cleared...
  goto eof
)

if '%1'=='-q' goto doit

if '%1' neq '' (
  echo.
  echo 	'%1' not recognised...
  goto eof
)

:doit
rem	Operating System			Version Number	Other Information
rem	Windows 1.0					1.04
rem	Windows 2.0					2.11
rem	Windows 3.0					3
rem	Windows NT 3.1				3.10.528
rem	Windows for Workgroups 		3.11		3.11
rem	Windows NT Workstation 		3.5			3.5.807
rem	Windows NT Workstation 		3.51		3.51.1057
rem	Windows 95					4.0.950
rem	Windows NT Workstation 		4.0			4.0.1381
rem	Windows 98					4.1.1998
rem	Windows 98 Second Edition	4.1.2222
rem	Windows Me					4.90.3000
rem	Windows 2000 Professional	5.0.2195
rem	Windows XP					5.1.2600	Current SP3
rem	Windows XP Professional x64 Edition		5.2.3790
rem	Windows Vista				6.0.6000	Current Version changed to 6.0.6002 with SP2
rem	Windows 7					6.1.7600	RTM
rem	Windows 8					6.2.9200
rem	Windows Server 2012 R2		6.3.(????)
rem	Windows 8.1					6.3.9600
rem	Windows 10					10.0.10586

rem	Check Windows Version...

for /f "usebackq tokens=*" %%i in (`ver`) do set w_vn=%%i
echo %w_vn% | findstr "5\.1\."	> nul
if %errorlevel% == 0 set m_vn=XP
echo %w_vn% | findstr "6\.1\."	> nul
if %errorlevel% == 0 set m_vn=W7
echo %w_vn% | findstr "6\.2\."	> nul
if %errorlevel% == 0 set m_vn=W8
echo %w_vn% | findstr "10\.0\."	> nul
if %errorlevel% == 0 set m_vn=WX

if not defined m_vn (
  echo.
  echo 	This version of %w_vn% may not be supported - let's see...
  echo.
)

set BCPLenv=Vn %vn%.%rvn%	%edt%

rem	Go to the directory hosting this script...
pushd "%~dp0"
rem	Find the directory name we're executing from:
for %%* in (.) do set BCPLdir=%%~nx*
if not %BCPLdir%==BCPLkit-0.9.7 (
  echo.
  echo 	Please:	a] 'cd' to the BCPLkit-0.9.7 directory and
  echo 		b] run %0 there
  echo.
  goto eof
)

for /f "usebackq tokens=*" %%i in (`cd`) do set pwd=%%i

rem	go to parent directory...
cd ..
for /f "usebackq tokens=*" %%i in (`cd`) do set BCPLroot=%%~pni
rem	for /f "usebackq tokens=*" %%i in (`cd`) do set BCPLroot=%%i


rem	parse the path, looking for the CodeBlocks gcc.exe installation:

for %%i in (gcc.exe) do set cbp="%%~dp$PATH:i"

if not exist %cbp% (
  echo.
  echo CodeBlocks MinGW path ^(%cbp%^) does not exist - aborting...
  echo.
  goto eof
)

if not '%cbp%' == '' (
  cd %cbp%\..
  for /f "usebackq tokens=*" %%i in (`cd`) do set GnuWin=%%~pni
  cd %pwd%
) else (
  echo.
  echo 	CodeBlocks MinGW not found - aborting...
  echo.
  goto eof
)
	
rem if exist "c:\program files\codeblocks\mingw" set GnuWin=C:\Program Files\CodeBlocks\MinGW
rem if exist "c:\program files (x86)\codeblocks\mingw" set GnuWin=C:\Program Files ^(x86^)\CodeBlocks\MinGW

rem if '%GnuWin%'=='' (
rem   echo.
rem   echo 	CodeBlocks MinGW not found - aborting...
rem   echo.
rem   goto eof
rem )

rem		The following 'set' commands specify:
rem			a)	BCPLsrc:  your BCPL compiler installation directory
rem			b)	BCPLcopt: your copt (code optimiser) installation directory
rem			c)	BCPLlib:  your BCPL library (loader object file) installation directory
rem			d)	BCPLhdrs: your BCPL header (GET file) installation directory
rem			e)	MCPLhdrs: your MCPL header (GET file) installation directory
rem			f)	BCPLocd:  your BCPL Ocode compiler installation directory
rem			g)	minGWVn:  your MinGW installation version number
rem			h)	GnuWin:   your MinGW (GNU for Windows) c compiler/assembler/loader installation directory
rem			i)	BCPLenv:  the version number of this script


setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

rem echo on
set pos=1
:loop
set /a pos+=1

for /f "usebackq tokens=1,%pos%" %%i in (`gcc --version`) do (
  if '%%i'=='gcc' (
    set item=%%j
rem echo '!item!'
	set ch=!item:~0,1!
	set cp=!ch!
	set /a ch+=0
rem		if 'ch' is numeric it will be unchanged...
    if '!ch!'=='!cp!' (
rem 	...therefore we assume that it's the version number
      echo set minGWVn=!item!>minGWVn.txt
rem 	(but because this is running 'setlocal' we can't 'set minGWVn=...' here
rem 	 ...it will disappear when we 'endlocal')
	  set flag=done
    )
  )
)
rem pause
if defined flag goto xloop
if %pos% gtr 10 (
  echo.
  echo 	Unable to find MinGW version!
  echo 	...you must 'set minGWVn=[version_number]' manually
  echo.
  goto xloop
)
goto loop

:xloop
endlocal

set BCPLsrc=%BCPLroot%\bcplkit-0.9.7\src
set BCPLcopt=%BCPLroot%\copt
set BCPLlib=%BCPLroot%\BCPL 2006-06-16
set BCPLhdrs=%BCPLroot%\include
set MCPLhdrs=%BCPLhdrs%
set BCPLdir=
if exist minGWVn.txt for /f "tokens=2" %%i in (minGWVn.txt) do (
  set %%i
  rem 	echo '%%i'
)
del /q/f minGWVn.txt

pushd
  cd ..\..
  for /f "usebackq tokens=*" %%i in (`cd`) do if exist BCPL-0.9.8 set BCPLocd=%%~pni\BCPL-0.9.8
popd

:envset

if '%1'=='-q' goto eof

echo.
echo 	BCPL environment established as:
echo.
set BCPL
set MCPL
set GnuWin
set minGWVn
echo.

:eof

set cbp=
set m_vn=
set w_vn=
set vn=
set rvn=
set edt=
popd
