/* Copyright (c) 2012 Dave Cannon.  All rights reserved. */

/* $Id: rt.c,v 1.1 2012/08/24 14:52:48 */

#include <stdio.h>
/*	FILE struct, as defined in stdio.h; comments from the K&R C Programming book; (FCB...) notes from rt.s
typedef struct _iobuf
{
	char*	_ptr;		// next character position:		(FCB.BF,0)
	int	_cnt;			// number of characters left:	(FCB.XP,4)
	char*	_base;		// location of buffer:			(FCB.CP,8)
	int	_flag;			// mode of file access:			(FCB.FD,12)
// Flags for the iobuf structure
#define	_IOREAD	1		// currently reading
#define	_IOWRT	2		// currently writing
#define	_IORW	0x0080	// opened as "r+w"
	int	_file;			// file descriptor(?):			(FCB.FL,16)
	int	_charbuf;		// (for use by ungetc()?)
	int	_bufsiz;		// (size of buffer?)
	char*	_tmpfname;	// (name of temporary file(?))
} FILE;
*/

#define DEBUG

#include <stdlib.h>
#include "blib.h"

#define ENDSTREAMCH -1

#define FD_STDIN 0
#define FD_STDOUT 1
#define FD_STDERR 2

#define FTSZ 20
#define FCBCNT 8						// (Why aren't FTSZ and FCBCNT the same?)
#define FCBSIZ sizeof(FILE)
#define BUFSIZ 512

#define FCB_FL 16

#define STACKBASE 54
#define STACKEND 55


static FILE *ft[FTSZ];
static int fi, fo;

int *G;

void
main (int argc, char *argv[])
{
	int pno = argc;
	char *pvec = *argv;
	
	#ifdef DEBUG
		int i;
		printf("\tSTART called, address of GLOBAL vector is %d\n", &G);
		for (i = 0; i < 100; i++)
			printf("\t\tG:%d = %d\n", i, &G[i]);
	#endif
	initbcpl(pno, pvec);
}


void
stacktrace(void)
{
	#ifdef DEBUG
		printf("\tstacktrace() called...\n");
	#endif
	printf("\t\tSTACKBASE (G:%d)=%d\tSTACKEND (G:%d)=%d\n", STACKBASE, G[STACKBASE*4], STACKEND, G[STACKEND*4]);
}


int
rtinit(void)
{
	int i;

/*	Original assembler established i/o buffers for first eight potential files...
	...we'll let the c library do the hard stuff
	
    char *eax;
	FILE *ebx;

    eax = malloc(FCBCNT*BUFSIZ);		// Grab memory for FCBCNT buffers

	for (i = 0; i < FCBCNT; i++) {
		ebx = ft[i];					// Collect pointer to FILE structure
		ebx->_base = eax;				// save address of BUFSIZ buffer in FILE._base
		ebx->_bufsiz = BUFSIZ;			// ...and its size in FILE._bufsiz
		eax = eax+BUFSIZ;				// calculate address of next BUFSIZ chunk
	}
*/
	#ifdef DEBUG
		printf("\trtinit() called...\n");
	#endif
	ft[FD_STDIN] = stdin;				// Save file pointer to 'stdin'...
	fi = FD_STDIN;						// pre-	selectinput(stdin)
	ft[FD_STDOUT] = stdout;				// ...and 'stdout'
	fo = FD_STDOUT;						// pre-	selectoutput(stdout)
	ft[FD_STDERR] = stderr;				// ...and 'stderr'
	fo = FD_STDERR;						// pre-	selectoutput(stderr)

	for (i = 3; i < FTSZ; i++)
		ft[i] = NULL;					// Clear remainder of pointers to NULL for ftslot()
	#ifdef DEBUG
	for (i = 0; i < FTSZ; i++)
		printf("\t\tft[%d] = %d\n", i, ft[i]);
	#endif
    return ft[FD_STDERR]->_file;		// Return last file descriptor created
}

void
rtexit(void)
{
	int i;

	#ifdef DEBUG
		printf("\trtexit() called...\n");
	#endif
	for (i = 3; i < FTSZ; i++) {
		if (ft[i] != NULL) {
			fclose(ft[i]);				// Close open file...
			ft[i] = NULL;				// ...clear FILE * pointer to NULL
		}
	}
}


char *
sbrk(int s)
{	// request a slice of dynamic memory:
	#ifdef DEBUG
		printf("\tmalloc() called requesting: %d bytes\n", s);
	#endif
	return malloc(s);
}

static char *
cstr(int s)
{
    char *st, *a;
    int n, i;

	#ifdef DEBUG
		printf("\tcstr() called...\n");
	#endif
	a = (char *)s;
    n = a[0];
    st = malloc(n + 1);
    for (i = 1; i <= n; i++)
        st[i - 1] = a[i];
    st[n] = 0;
	#ifdef DEBUG
		printf("\tcstr() text is: %s\n", st);
	#endif
    return st;
}

static int
ftslot(void)
{
    int i;

	#ifdef DEBUG
		printf("\tftslot() called...", i);
	#endif
    for (i = 2; i < FTSZ; i++)
        if (ft[i] == NULL) {
			#ifdef DEBUG
				printf(" returning: %d\n", i);
			#endif
            return i;
		}
    return -1;
}

int
findinput(int s)
{
    char *st = cstr(s);
    int x;

	#ifdef DEBUG
		printf("\tfindinput(%s) called\n", st);
	#endif
    x = ftslot();
    if (x != -1) {
        ft[x] = fopen(st, "r");
        if (ft[x] == NULL)
            x = -1;
    }
    free(st);
    return x;
}

int
findoutput(int s)
{
    char *st = cstr(s);
    int x;

	#ifdef DEBUG
		printf("\tfindoutput(%s) called\n", st);
	#endif
    x = ftslot();
    if (x != -1) {
        ft[x] = fopen(st, "w");
        if (ft[x] == NULL)
            x = -1;
    }
    free(st);
    return x;
}

void
wrch(int c)
{
	#ifdef DEBUG
		printf("\twrch(%c) called\n", c);
	#endif
    fputc(c, ft[fo]);
}

int
rdch(void)
{
    int c;
	
	c = fgetc(ft[fi]);
	#ifdef DEBUG
		printf("\trdch(%c) called, returning\n", c);
		stacktrace();
	#endif
	if (c == EOF) 
		c = ENDSTREAMCH;
	return c;
}

int
unrdch(void)
{
	#ifdef DEBUG
		printf("\tunrdch() called\n");
	#endif
	return ungetc(ft[fi]->_charbuf, ft[fi]);
}

void
selectinput(int x)
{
    fi = x;
	#ifdef DEBUG
		printf("\tselectinput(%d) called\n", x);
	#endif
}

void
selectoutput(int x)
{
    fo = x;
	#ifdef DEBUG
		printf("\tselectoutput(%d) called\n", x);
	#endif
}

int
input(void)
{
	#ifdef DEBUG
		printf("\tinput() called, returning: %d\n", fi);
	#endif
    return fi;
}

int
output(void)
{
	#ifdef DEBUG
		printf("\toutput() called, returning: %d\n", fo);
	#endif
    return fo;
}

void
endread(void)
{
	#ifdef DEBUG
		printf("\tendread() called...\n");
	#endif
    if (fi > 1) {
        fclose(ft[fi]);
        ft[fi] = NULL;
    }
    fi = 0;
}

void
endwrite(void)
{
 	#ifdef DEBUG
		printf("\tendwrite() called...\n");
	#endif
   if (fo > 1) {
        fclose(ft[fo]);
        ft[fo] = NULL;
    } else
        fflush(ft[fo]);
    fo = 1;
}

int
rwd_bcpl(int x)
{
	#ifdef DEBUG
		printf("\trewind() called...\n");
	#endif
	rewind(ft[x]);
}

int
getvec(int s)
{
	#ifdef DEBUG
		printf("\tgetvec(%c) called\n", s);
	#endif
    return (int) malloc(s);
}

void
freevec(int v)
{
	#ifdef DEBUG
		printf("\tfreevec(%c) called\n", v);
	#endif
    free((void *) v);
}

float
flop(float a, int b)
{
	float c;
	
//	c = (int) a;
	c = (float) b;
	return c;
}

