
// File ML1HDR

// Header file for ML/I macro processor on EMAS 2900

// BCPL version by R.D. Eager   University of Kent   MCMLXXXV


GET "LIBHDR"

MANIFEST $(   // Tunable MANIFESTs
lhv            = 32    // Size of hash table
texmax         = 64    // Maximum length of construction in an error message
htmax          = texmax/2 - 4
pvarnum        = 10    // Initial number of permanent variables
svarnum        = 23    // Number of system variables
error_quota    = 500   // Default lines limit on debugging file
error_character= '?'   // Substituted for illegal input characters

debugging_     = FALSE // Set to TRUE during development and testing
$)

MANIFEST $(   // Miscellaneous constants
zeropt         = 0     // Must be less than or equal to any pointer value
nullpt         = 0     // Must be different from any pointer value

sdbsz          = 18    // Size of scanning description block
opdbsz         = 9     // Size of operation macro description block
allsz          = 3     // Size of the OPT-ALL block
edbsz          = 8     // Size of the error description block
nodesz         = 3     // Size of a node entry on backwards stack

opmk           = 0     // Operation macro
locmk          = 1     // Local NEC macro
uinsmk         = 2     // Unprotected insert
pinsmk         = 3     // Protected insert
strmk          = 4     // Straight-scan macro

endchn  = maxint       // Closing delimiter/end of chain marker
exclmk  = maxint - 1   // Exclusive delimiter marker
withmk  = maxint - 2   // For WITH keyword
wthsmk  = maxint - 3   // For WITHS keyword
spcsmk  = maxint - 4   // For SPACES keyword
slrep   = maxint - 5   // For SL keyword
$)

MANIFEST $(   // Codes for operation macros
mc_nodef       = 0     // Suppress local macro definitions
mc_nowarn      = 1     // Suppress local warning marker definitions
mc_noins       = 2     // Suppress local insert definitions
mc_noskip      = 3     // Suppress local skip definitions
mc_def         = 4     // Define a macro (local or global)
mc_warn        = 5     // Define a warning marker (local or global)
mc_ins         = 6     // Define an insert (local or global)
mc_skip        = 7     // Define a skip (local or global)
mc_set         = 8     // Update a macro-time variable
mc_go          = 9     // Perform a macro-time goto
mc_alter       = 10    // Alter name of secondary delimiter, etc.
mc_note        = 11    // Macro-time debugging aid
mc_pvar        = 12    // Alter number of permanent variables
mc_leng        = 13    // Find length of macro-time string
mc_sub         = 14    // Extract macro-time substring
mc_stop        = 15    // Define a stop marker
$)

MANIFEST $(   // Values for 'bestpl'
bp_copy        = 0     // Copy atom to output text
bp_macro       = 1     // Macro name encountered
bp_warn        = 2     // Warning marker encountered
bp_insert      = 3     // Insert name encountered
bp_skip        = 4     // Skip name encountered
bp_sdelim      = 5     // Scanning for secondary delimiter
bp_null        = 6     // Produce no output text
bp_awarn       = 7     // After warning marker encountered
bp_stack       = 8     // Place atom on forwards stack
bp_stop        = 9     // Stop marker encountered
$)

MANIFEST $(   // Types of construction
ty_stop        = 0     // Stop marker
ty_macro       = 1     // Macro
ty_warn        = 2     // Warning marker
ty_insert      = 3     // Insert
ty_skip        = 4     // Skip
$)

MANIFEST $(   // Types of keyword in structure representations
kw_with        = 0     // WITH
kw_withs       = 1     // WITHS
kw_opt         = 2     // OPT
kw_or          = 3     // OR
kw_all         = 4     // ALL
$)

MANIFEST $(   // Values for 'dbugsw'
db_source      = 0     // Scanning source text
db_repl        = 1     // Scanning replacement text
db_oparg       = 2     // Scanning for operation macro argument
db_roparg      = 3     // Scanning text of operation macro argument or insert
db_subarg      = 4     // Scanning for substitution macro argument
db_eval        = 5     // Scanning operation macro or first insert
db_delim       = 6     // Scanning for delimiter
$)

MANIFEST $(   // Global assignments
gvb            = firstfreeglobal   // Base of variables
gpb            = gvb+111           // Base of procedures
$)

GLOBAL $(   // Variables

// Scanning description block (sdb)

argct          : gvb+000   // Counts number of arguments when nested construction is scanned
stakpt         : gvb+001   // Points at latest SDB on stack ('nullpt' whilst scanning source text)        \
argpt          : gvb+002   // Points at argument vector (scanning replacement text)                       |
dbugpt         : gvb+003   // Points at orlink preceding macro name                                       |
spt            : gvb+004   // Points at last scanned character                                            |  edb - error
mchlin         : gvb+005   // Set to current value of 'linect' when a nested construction is encountered  |  description block
linect         : gvb+006   // Line count of current text                                                  |
argno          : gvb+007   // Number of argument or delimiter                                             |
dbugsw         : gvb+008   // Indicates state of scan                                                     /
tvarpt         : gvb+009   // Points at temporary variables
hashpt         : gvb+010   // Points at local hash table
mtchpt         : gvb+011   // Points at orlink when a nested construction is encountered
stoppt         : gvb+012   // Points one beyond last character of current text
labpt          : gvb+013   // Head of chain of labels
inffpt         : gvb+014   // Points at start of source text on forwards stack
skval          : gvb+015   // Number of designated label for forward MCGO
sklin          : gvb+016   // Line number in which MCGO occurred
ohsw           : gvb+017   // TRUE iff a new hash table has been stacked for this level

// Operation macro description block (opdb)

topspt         : gvb+018   // Points at latest 'opdb' on backwards stack
mhshpt         : gvb+019   // Value of 'hashpt' when macro was called
sqpt           : gvb+020   // Safe variable, miscellaneous uses
linkpt         : gvb+021   // Link for 'stkarg'
arglen         : gvb+022   // Length of value of current argument
optyp          : gvb+023   // Type, e.g. global, local, insert. Also other uses
sqnum          : gvb+024   // Safe variable, miscellaneous uses
sqsw           : gvb+025   // Safe variable, miscellaneous uses
ntypsw         : gvb+026   // Type of construction being defined

// Variables used in processing OPT - ALL brackets

allpt          : gvb+027   // Head of chain of nextlinks to be attached to delimiter following ALL
optpt          : gvb+028   // Last entry on orlink chain
opthpt         : gvb+029   // Head of orlink chain

// Variables requiring dynamic initialisation (dib)

endpt          : gvb+030   // Points at end of backwards stack
pvarpt         : gvb+031   // Points at permanent variables
lfpt           : gvb+032   // Points at last used location on backwards stack
pvnum          : gvb+033   // Number of permanent variables
glbwsw         : gvb+034   // Global warning switch

// Miscellaneous variables

knpt           : gvb+035   // Points to node marker entry in delimiter chain
svarpt         : gvb+036   // Points at system variables
bfndpt         : gvb+037   // Best-so-far value of 'fndpt'
infopt         : gvb+038   // Points beyond currently matched LID
binfpt         : gvb+039   // Best-so-far value of 'infopt'
ffpt           : gvb+040   // Points to first free location on forwards stack
tempt          : gvb+041   // Temporary
chanpt         : gvb+042   // Used in chaining
eriapt         : gvb+043   // Points at value of operation macro argument
bespt          : gvb+044   // Best-so-far value of 'spt'
htabpt         : gvb+045   // Points at current hash table
idpt           : gvb+046   // Points at current identifier
cllfpt         : gvb+047   // Points at top entry after scanning information is stacked
nargpt         : gvb+048   // Points at argument vector + 1
stffpt         : gvb+049   // Points at start of global macros
delpt          : gvb+050   // Head of chain of delimiters searching for
skiplv         : gvb+051   // Level of skip nesting
nestlv         : gvb+052   // Nesting level of calls and skips during scanning
oplev          : gvb+053   // Level of operation macros and inserts
bindic         : gvb+054   // Best-so-far value of 'indic'
levl           : gvb+055   // Level of macros and inserts
invoct         : gvb+056   // Count of macro calls
indic          : gvb+057   // Contents of nextlink
chlink         : gvb+058   // Used in chaining
idlen          : gvb+059   // Length of current identifier
bestpl         : gvb+060   // Switch value used in basic scan routine
offset         : gvb+061   // Offset in hash table
meval          : gvb+062   // Miscellaneous; used for numerical values calculated at macro time
beslin         : gvb+063   // Best-so-far value of 'linect'
caltyp         : gvb+064   // First number in information block
tlinct         : gvb+065   // Temporary storage for 'linect'
masksw         : gvb+066   // Used to indicate which types of construction are recognised
coptsw         : gvb+067   // For skips; reflects setting of text option
copdsw         : gvb+068   // For skips; reflects setting of delimiter option

// Variables used in processing structure representations

nodept         : gvb+069   // Head of chain of links to be attached to next delimiter
nnodpt         : gvb+070   // Points at current node entry on backwards stack
lnodpt         : gvb+071   // Points at topmost node entry on backwards stack
ollfpt         : gvb+072   // Previous value of 'lfpt'
ndefpt         : gvb+073   // Destination of newly defined construction
delct          : gvb+074   // Count of delimiters
optlev         : gvb+075   // Level of OPT-ALL brackets
nodesw         : gvb+076   // For syntax checking ('evtree' and 'getdel')
keysw          : gvb+077   // For syntax checking ('evtree' and 'getdel')
consw          : gvb+078   // For syntax checking ('evtree' and 'getdel')
exitsw         : gvb+079   // For syntax checking ('evtree')

// Variables used in processing macro expressions or inserts

varpt          : gvb+080   // Points at vector of variables
$)

GLOBAL $(   // Names of table items - dynamically initialised
ghshtb         : gvb+081   // Global hash table
knrep          : gvb+082   // Points at current node marker
laychn         : gvb+083   // Chain of names of layout characters
kspacs         : gvb+084   // Points at SPACES keyword
keychn         : gvb+085   // Head of keyword chain
delchn         : gvb+086   // Chain of secondary delimiters
dif            : gvb+087   // Head of IF/UNLESS chain for MC-GO
dge            : gvb+088   // Head of relation chain for MC-GO
das            : gvb+089   // Head of AS/SSAS chain for MC-DEF
spcsrp         : gvb+090   // Points at representation of SPACE
$)

GLOBAL $(   // Miscellaneous variables
at_edb         : gvb+091   // \
at_sdb         : gvb+092   //  |__ Addresses of blocks of scanning data
at_opdb        : gvb+093   //  |
at_all         : gvb+094   // /
at_s1          : gvb+095   // Address of S1; for efficient access
at_s2          : gvb+096   // Address of S2; for efficient access
at_s5          : gvb+097   // Address of S5; for efficient access
typvec         : gvb+098   // Work area for 'prtype'
convarea       : gvb+099   // Work area for 'mdconv'
nlsw_          : gvb+100   // TRUE iff last character read was a newline
scanlevel      : gvb+101   // Level of scanning closure
$)

GLOBAL $(   // Machine-dependent variables
at_s10         : gvb+102   // \
at_s12         : gvb+103   //  |
at_s16         : gvb+104   //  |
at_s17         : gvb+105   //  |
at_s19         : gvb+106   //  |-- Addresses of S-variables; for efficient access
at_s20         : gvb+107   //  |
at_s21         : gvb+108   //  |
at_s22         : gvb+109   //  |
at_s23         : gvb+110   // /
$)

GLOBAL $(   // Internal routines   (alphabetical order)
advnce         : gpb+000
basic_scan     : gpb+001
bsnext         : gpb+002
bstrex         : gpb+003
chatom         : gpb+004
chekid         : gpb+005
ckvaly         : gpb+006
cmpare         : gpb+007
decalv         : gpb+008
encall         : gpb+009
entext         : gpb+010
er1tst         : gpb+011
erlia          : gpb+012
erlmd          : gpb+013
erlme          : gpb+014
erlovf         : gpb+015
erlso          : gpb+016
ermtst         : gpb+017
ersic          : gpb+018
ersnw          : gpb+019
evexit         : gpb+020
evopt          : gpb+021
evor           : gpb+022
gargch         : gpb+023
getdel         : gpb+024
getexp         : gpb+025
gmeadd         : gpb+026
gsatom         : gpb+027
gtatom         : gpb+028
init_tables    : gpb+029
joinch         : gpb+030
ludel          : gpb+031
lulayk         : gpb+032
macerr         : gpb+033
mcabrt         : gpb+034
mihalt         : gpb+035
milogic        : gpb+036
mkroom         : gpb+037
opexit         : gpb+038
prarg          : gpb+039
prctxt         : gpb+040
prenv          : gpb+041
prerr          : gpb+042
prlid          : gpb+043
prmiss         : gpb+044
prname         : gpb+045
prscan         : gpb+046
prtype         : gpb+047
prviz          : gpb+048
ressp          : gpb+049
sbstpl         : gpb+050
setpts         : gpb+051
setype         : gpb+052
sklab          : gpb+053
stkarg         : gpb+054
stkhsh         : gpb+055
subchk         : gpb+056
tebest         : gpb+057
tesdel         : gpb+058
tespac         : gpb+059
tewith         : gpb+060
unopdb         : gpb+061
unsdb          : gpb+062
$)

GLOBAL $(   // Machine-dependent routines
bmove          : gpb+063   // Moves blocks of data, starting at their end
bumpff         : gpb+064   // Increases 'ffpt', checking for stack overflow
declf          : gpb+065   // Decreases 'lfpt', checking for stack overflow
fmove          : gpb+066   // Moves blocks of data, starting at their start
mdconv         : gpb+067   // Converts numbers to strings
mddiv          : gpb+068   // Performs division
mderpr         : gpb+069   // Outputs messages to the debugging file
mdfinal        : gpb+070   // Performs achine-dependent finalisation
mdfind         : gpb+071   // The hashing function
mdinit         : gpb+072   // Machine-independent initialisation
mdnum          : gpb+073   // Converts strings to numbers
mdouch         : gpb+074   // Main output routine
mdread         : gpb+075   // Main input routine
$)

// End of file ML1HDR
